

/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Lucas Silveira
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
# include <omp.h>
#include <math.h>
#include <string.h>
#include "fitness.c"
#include "structure.h"
#include "GerarPopuInicial.c"

float prob_crossover;     //Crossover Probability
float porc_selection;     //% of the Population that will be selected for Crossover
float porc_reemp;           //Pct. of the Population from where it will be performed
float prob_mutation;     //Mutation Probability
int * identity;
int size;
int sentinela;



#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000


float avaliaPopulation(population *pob)
{
    float *vet = NULL;
    float soma = 0.0, media=0.0, var = 0.0;
    int i;
    vet = (float *)malloc(pob->sizePopulation * sizeof(float));
    if(vet == NULL)
    {
        printf("\n!!!ERRO!!! Fracasso na alocacao do vetor.\n");
        exit(1);
    }
    
    for(i = 0; i < pob->sizePopulation; i++)
    {
        vet[i]=pob->genome[i].fitness;
        soma = soma + vet[i];
        //printf("%f\n",soma);
    }
    media = (soma)/pob->sizePopulation;
    for(i = 0; i < pob->sizePopulation; i++)
    {
        /*var(x) = Sn ((xi - média)^2 ) /N*/
        var = (var + (pow(vet[i] - media,2))/pob->sizePopulation);
        /*s(x) = Sn ((xi - média)^2) / (N - 1)*/
    }
    // dp = sqrt(var);
    free(vet); /*libera o vetor alocado anteriormente*/
    return (media+var);
}



void substitution(population *pob){
    int i,j,pos_reemp;
    int conta=0;
    /*descendants replace the current Population*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        //seek position of substitution
        int tamanhoReemp = pob->topPopulation - pob->baseReplace;
        pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(values between 0 and tamanhoReemp-1)+baseReplace
        
        //apply substitution
        if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
            //copy fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copy genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
    }
    
}


/*
 Dado um vetor v e três inteiros i, m e f, sendo v[i..m] e v[m+1..f] vetores ordenados,
 coloca os elementos destes vetores, em ordem crescente, no vetor em v[i..f].
 */
void Merge(float *v, int *v1, float *c, int *c1, int i, int m, int f) {
    int z,t,w,
    iv = i, ic = m + 1;
    
    for (z = i; z <= f; z++)
    {
        c[z] = v[z];
        c1[z]=v1[z];
    }
    
    z = i;
    
    while (iv <= m && ic <= f) {
        /* Invariante: v[i..z] possui os valores de v[iv..m] e v[ic..f] em ordem crescente. */
        
        if (c[iv] < c[ic])
        {
            t=z++;
            w=iv++;
            v[t] = c[w];
            v1[t]=c1[w];
        }
        else /* if (c[iv] > c[ic]) */
        {
            t=z++;
            w=ic++;
            v[t] = c[w];
            v1[t]=c1[w];
        }
    }
    
    while (iv <= m)
    {   t=z++;
        w=iv++;
        v[t] = c[w];
        v1[t] = c1[w];
    }
    
    while (ic <= f)
    {   t=z++;
        w=ic++;
        v[t] = c[w];
        v1[t] = c1[w];
    }
}


/*
 Dado um vetor de inteiros v e dois inteiros i e f, ordena o vetor v[i..f] em ordem crescente.
 O vetor c é utilizado internamente durante a ordenação.
 */
void Sort(float *v,int *v1, float *c, int *c1, int i, int f) {
    if (i >= f) return;
    
    int m = (i + f) / 2;
    
    Sort(v, v1, c,c1, i, m);
    Sort(v, v1, c, c1, m + 1, f);
    
    /* Se v[m] <= v[m + 1], então v[i..f] já está ordenado. */
    if (v[m] <= v[m + 1]) return;
    
    Merge(v, v1, c,c1, i, m, f);
}

/*
 Dado um vetor de inteiros v e um inteiro n >= 0, ordena o vetor v[0..n-1] em ordem crescente.
 */
void MergeSort(float *v,int *v1, int n) {
    float *c =  (float*)malloc(sizeof(float) * n);
    int *c1 = (int*)malloc(sizeof(int) * n);
    Sort(v,v1, c, c1, 0, n - 1);
    free(c);
    free(c1);
}


void countingSort(population *pob, int k){
    int i,j;
    int *c;//c[k];
    Genome *a;
    
    c = (int*)malloc(k*sizeof(int));
    
    a = (Genome*)malloc(pob->sizePopulation*sizeof(Genome));
    if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = (int*)malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
    }
    
    /*counting sort*/
    for(i=0; i<k; i++){
        c[i] = 0;
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        c[a[i].fitness] = c[a[i].fitness] + 1;
    }
    
    for(i=1; i<k; i++){
        c[i] = c[i] + c[i-1];
    }
    
    for(i=pob->sizePopulation-1; i>=0; i--){
        
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
        c[a[i].fitness] = c[a[i].fitness] - 1;
    }
    
    
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
    free(a);
    free(c);
}


//fiz ate aqui
void mutation(population *pob){
    int i,j;
    float prob;
    /*mutation about descendents, baseDescendence=0*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        for(j=0; j< size; j++)
        {
            prob = ((rand() % 100) + 1)/1000.0; //de 0.01, 0.02 ... a 1.0
            
            if(prob <= prob_mutation)
            {
                pob->genome[i].genes[j] = (rand() % size);
                //applied a gene mutation (flip sign)
//                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
//                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
            }
            
        }
        
    }//end for
}



//takes best solution
void selection(population *pob,  int n_generetion, int * better_soluction){
    
    countingSort(pob,pob->genome[0].size+2);
    
    if (pob->genome[0].fitness < *better_soluction)
        *better_soluction = pob->genome[0].fitness;
    
}

void crossover(population *pob) {
    int i,j,k,pos_pai1, pos_pai2, point_crossover;
    float prob;
    
    k = pob->baseDescendencia;
    
    for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
        if (k+1 == pob->sizeTotal)
            break;
        
        
        pos_pai1 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        
        for(j=0; j<10; j++){
            if (pos_pai1 != pos_pai2)
                break;
            pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        }
        
        prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
        if (prob <= prob_crossover)
        {//crossover
            point_crossover = rand() % size;//valores entre 0 e tamanho de permutation - 1
            //aplicar single_point_crossover na primera parte antes do ponto de crossover
            for(j=0; j<point_crossover; j++){
                pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            //aplicar single_point_crossover na segunda parte depois do ponto de crossover
            for(j=point_crossover; j<size; j++){
                pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            k = k + 2;
        }//end if crossover
    }//end for iniit
    pob->topeDescendencia = k;//update tope descendence
}




//populacao com 12 processadores
void Population_init(population *pob, int numberprocessadores){
    int i,j,n;
    
    //Calc size of Population
    n = size; //numero de rainhas
    pob->sizePopulation =  (int) (n * (log(n) / log(2))); //size of Population = (nlogn); log in base 2
    pob->sizePopulation =(2*pob->sizePopulation) * numberprocessadores;
    pob->basePopulation = 0;
    pob->topPopulation = pob->sizePopulation;
    
    pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de Population what go to crossover
    pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a the from of where will replace
    
    pob->baseDescendencia = pob->sizePopulation;
    pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
    pob->sizeTotal = pob->topeDescendencia;
    
    pob->genome = (Genome*)malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = (int*)malloc(n * sizeof(int));//alocar memoria para cada cromosoma
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in Population_init(x,x) in Alg_Geneti_simple\n");
        pob->genome[i].size = size;
    }
    
    /*create Population a partir of  permutacion*/
    // A genom is see with a permutation of concatenation chroms
    
    for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        for (j = 0; j < pob->genome[i].size; j++) {
            int r = rand() % size;
            pob->genome[i].genes[j] = r;
        }
        pob->genome[i].size = size;
    }//end for extern
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
        pob->genome[i].size = size;
    }
    
    
}//end create init Population



void fitness_Population(population *pob, int flag)
{
    //printf("-------------------------------- : %d \n", size);
   

    int i;
    if (flag > 1){ //fitness over descendants
//        #pragma omp parallel for num_threads(2) private (i)
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
        {
            
            pob->genome[i].fitness = calculateCostFunction(pob->genome[i].genes, pob->genome[i].size);
        }
    }
    else
    { // fitness over population
//        #pragma omp parallel for num_threads(2) private (i)
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
        {
           pob->genome[i].fitness = calculateCostFunction(pob->genome[i].genes, pob->genome[i].size);
        }
    }
}





//flag em uma roda vai comecar no inicio outro no final
int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, population *pob, int flag){
    int i, indice = 0;
    int porcento=sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: // pior
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: //aleatorio
            indice=(rand() % sizePopulation);
            break;
        case 3://semelhante
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->genome[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->genome[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    
    return indice;
    
    
}// end function Imigracao


//busca os indicies a ser enviado para outra ilha
int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= (int*)malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");
    switch (typeIndividual) {
        case 1://melhor
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: // pior
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3: //aleatorio
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=(rand() % sizePopulation);
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, population *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    msgGenome enviar,recebe;
    
    // If we are the root process, send our best data to everyone
    enviar.size=pob->genome[0].size;
    enviar.fitness=pob->genome[0].fitness;
    for (i=0; i<pob->genome[0].size; i++)
        enviar.genes[i]=pob->genome[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(msgGenome),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(msgGenome),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->sizePopulation; j++)
                {
                    if (recebe.fitness < pob->genome[j].fitness)
                    {
                        pob->genome[j].fitness=recebe.fitness;
                        for (i=0; i<pob->genome[j].size; i++)
                        {
                            pob->genome[j].genes[i]=recebe.genes[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function


//return 1 to insert and 0 to discard
// Pgene  vai de 0.1 ate 1
// uma flag sera imposta a fim de configura a quantidade da populacao a ser comparada
int CompareIndividuals(population *pob, msgGenome genomaAlvo, float Pgenes) {
    int i,j;
    int flag=1;
    int contador=0;
    int intervalo=pob->sizePopulation * Pgenes;
    //printf("intervalo levado em consideracao %d sobre %d\n ", intervalo, pob->sizePopulation);
    for (i=0; i<intervalo; i++) {
        
        for (j=0; j<pob->genome[i].size; j++)
        {
            if(pob->genome[i].genes[j]==genomaAlvo.genes[j])
            {    contador++;
            }
        }//end for
        
        if (contador==pob->genome[i].size) {
            flag=0;
            break;
        }
        contador=0;
    }//end for
    
    return flag;
}





//implementa topologia full graph
//flag é pra marcar se for igual a 1 significa q esta terminando as geracoes e todos as ilhas necessitas saber os resultados
void my_full(MPI_Comm communicator, int world_rank, population *pob, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes) {

    int world_size, j,i,ii,iii,error_code, index;
    float prob;
    int envio=-1;
    int flag=0;
    int receptor=-1;
    int avaliador=1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    msgGenome enviar,recebe;
    
    //buscar indices para ser enviado
    
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->sizePopulation);
    
    
    
    for (i=0; i< qtdIndividual; i++)
    {
        // If we are the root process, send our data to everyone
        enviar.size=pob->genome[IndicesEnvia[i]].size;
        enviar.fitness=pob->genome[IndicesEnvia[i]].fitness;
        for (j=0; j<pob->genome[IndicesEnvia[i]].size; j++)
            enviar.genes[j]=pob->genome[IndicesEnvia[i]].genes[j];
        
        //sent to world_rank+1
        
        if (world_rank!=0)
            
            
            
        {
            
            int error_code= MPI_Send(&enviar,//the data
                                     
                                     sizeof(msgGenome),
                                     
                                     MPI_BYTE,
                                     
                                     0,//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            if (error_code != MPI_SUCCESS)
                
            {
                
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
                
                
                
            }
            
            
            
        }
        
        
        
        else
            
        {
            
            for( iii = 0; iii < world_size; iii++)
                
            {
                
                
                
                
                
                if (iii != world_rank)
                    
                {
                    
                    int error_code=MPI_Recv(&recebe,
                                            
                                            sizeof(msgGenome),
                                            
                                            MPI_BYTE,
                                            
                                            iii, // processo que enviou
                                            
                                            0,//tag
                                            
                                            communicator,
                                            
                                            MPI_STATUS_IGNORE);
                    
                    
                    
                    if (error_code != MPI_SUCCESS)
                        
                    {
                        
                        char error_string[BUFSIZ];
                        
                        int length_of_error_string;
                        
                        MPI_Error_string(error_code, error_string, &length_of_error_string);
                        
                        printf("%3d: %s\n", world_rank, error_string);
                        
                    }//end if
                    
                    
                    
                    if (flag) {
                        
                        flag=0;
                        
                    }
                    
                    else
                        
                        flag=1;
                    
                    
                    
                    index=Imigracao( imigracao, pob->sizePopulation, recebe.fitness,  pob,  flag);
                    //printf("index %d tamanho total %d\n", index,pob->sizePopulation);
                    if (index>=pob->sizePopulation) {
                        index=pob->sizePopulation-1;
                        printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
                    }
                    
                   
                    avaliador=CompareIndividuals(pob, recebe, Pgenes);
                    if (avaliador)
                    {
                        pob->genome[index].fitness=recebe.fitness;
                        
                        for (ii=0; ii<pob->genome[index].size; ii++)
                            pob->genome[index].genes[ii]=recebe.genes[ii];
                    }

                    
                }//end if
                
            }//end for

        }//else
        
    }//end for
    
    free(IndicesEnvia);
    



}//end function anel









void Genetic(MPI_Comm communicator,int world_rank, float prob_crossover, float porcentaje_selec,float porcentaje_reemp, float prob_mutation,int emigracao,int typeIndividual,int qtdIndividual,int imigracao,float intervaloMigracao, float Pgenes, int queens) 
{
    permutation perm;
    sentinela=0;
    population pob, pob2;
    int better_soluction,nro_generacion,i,num_generation, world_size;
    MPI_Comm_size(communicator, &world_size);
//    static const std::string constString("GENOME");
//    identity=read_permutation_of_file(constString.c_str(), &perm);
    size = queens;
   


    if (world_rank==0) {
        Population_init(&pob2, world_size);//crear poblacion inicial
        construirPopInit(world_size,&pob2, size, &perm);
        for(i=0;i<pob2.sizeTotal;i++){
            free(pob2.genome[i].genes);
        }
        free(pob2.genome);
    }

    
      MPI_Barrier(MPI_COMM_WORLD);
    
      //aqui cada processo carrega tua populacao
       CarregarPopInitial(world_rank, porcentaje_reemp, porcentaje_selec, &pob,&perm);
        fitness_Population(&pob, 1);
        better_soluction = pob.genome[0].fitness;
          num_generation = 100;
        int contador = num_generation * intervaloMigracao;
        for(nro_generacion=2; nro_generacion<= num_generation; nro_generacion++)
        {  //  printf("%d\n", nro_generacion);
            selection(&pob,nro_generacion,&better_soluction);//seleccion:sorting population
            crossover(&pob);
            mutation(&pob);
            fitness_Population(&pob,nro_generacion);
            substitution(&pob);
            contador--;

	

            if (contador==0)
            {

                my_full(communicator, world_rank, &pob, emigracao, qtdIndividual,typeIndividual, imigracao, Pgenes);
                contador = num_generation * intervaloMigracao;
            }
            
            if (nro_generacion==num_generation)
            {
                sendALL(communicator, world_rank, &pob);
                
            }//end if
            
            MPI_Barrier(MPI_COMM_WORLD);
            
        }
                
        if (world_rank==0) {
            selection(&pob,nro_generacion,&better_soluction);//seleccion:ordena populacao
            printf("%d", pob.genome[0].fitness);

	
        }

        for(i=0;i<pob.sizeTotal;i++){
            free(pob.genome[i].genes);
        }
        free(pob.genome);
        free(identity);
        
//    }
//    else
//        printf("*****************Input is not COTAIL************************");
   
    
    
    
}

int main(int argc,char **argv)
{
//    int a = 0;
//    scanf("%d", &a);
   
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    int seed2=atoi(argv[1]);
    
    /*
     float prob_crossover=atof(argv[2]);  //Crossover Probability
     float porcentaje_selec=atof(argv[3]);  //% of the population that will be selected for Crossover
     float porcentaje_reemp=atof(argv[4]); //Pct. of the population from where it will be performed
     float prob_mutation=atof(argv[5]);      //Mutation Probability
     int emigracao=atoi(argv[6]);// 1 para copiar e 2 para remover da populacao
     int typeIndividual=atoi(argv[7]); //1 melhor, 2 pior, 3 aleatorio
     int qtdIndividual=atoi(argv[8]); //1,2,3,4
     int imigracao=atoi(argv[9]); //1 pior ind., 2 ind. aleatorio 3. ind. muito semelhante
     float intervaloMigracao=atof(argv[10]); // 1 por cento ate 10
     // float Pgenes=atof(argv[11]); // 1 por cento ate 10
     */
     prob_crossover=atof(argv[2]);  //Crossover Probability
     porc_selection=atof(argv[3]);  //% of the population that will be selected for Crossover
     porc_reemp=atof(argv[4]); //Pct. of the population from where it will be performed
     prob_mutation=atof(argv[5]);      //Mutation Probability
    int emigracao=atoi(argv[6]);// 1 para copiar e 2 para remover da populacao
    int typeIndividual=atoi(argv[7]); //1 melhor, 2 pior, 3 aleatorio
    int qtdIndividual=atoi(argv[8]); //1,2,3,4
    int imigracao=atoi(argv[9]); //1 pior ind., 2 ind. aleatorio 3. ind. muito semelhante
    float intervaloMigracao=atof(argv[10]); // 1 por cento ate 10
    float Pgenes=atof(argv[11]); // 1 por cento ate 10
    int   queens= atoi(argv[12]);
    int world_rank;
    MPI_Init(NULL, NULL);
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed[world_rank]+seed2);//initialize random generator
    Genetic(MPI_COMM_WORLD, world_rank, prob_crossover,porc_selection,porc_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao, Pgenes, queens);
    MPI_Finalize();
    return 0;
}


