#ifndef ICOSTCALCULATOR_H
#define ICOSTCALCULATOR_H

#include <string>

/*!
 * \brief The \c ICostCalculator class is an interface for CostCalculator classes.
 */
class ICostCalculator
{
public:
    /*!
     * \brief A Virtual Destructor.
     */
    virtual ~ICostCalculator()
    {
    }

    /*!
     * \brief Getter to the the cost function name.
     * \return A string containing the cost function name.
     */
    virtual std::string getName() const = 0;

    /*!
     * \brief Getter to the number of objectives returned by this cost function.
     * \return The number of objectives calculated by this cost function
     */
    virtual unsigned getNumberOfObjectives() const = 0;

    /*!
     * \brief Getter to the default or suggested number of dimensions for this cost function.
     */
    virtual unsigned getDimensions() const = 0;
};

#endif // ICOSTCALCULATOR_H
