#include "application.h"

#include "applicationgenerator.h"

#include <iostream>
#include <sstream>
#include <string>

void Application::_populateTasksAndFlowsFromCsvFile(const std::string &fileName, const std::string &dirPath)
{
    _fileName = fileName;
    ApplicationGenerator appGenerator;

    appGenerator.generateTasksAndFlowsFromCsvFile(dirPath,
                                                  fileName,
                                                  _tasks,
                                                  _flows);
}

Application::Application(const std::string &fileName)
{
    _populateTasksAndFlowsFromCsvFile(fileName);
}

Application::~Application()
{
    for (std::pair<std::string, std::shared_ptr<Flow>> flow : _flows) {
        flow.second = nullptr;
    }
    _flows.clear();

    for (std::pair<std::string, std::shared_ptr<Task>> task : _tasks) {
        task.second = nullptr;
    }
    _tasks.clear();
}

Application::Application(const std::string &dirPath, const std::string &fileName)
{
    _populateTasksAndFlowsFromCsvFile(dirPath, fileName);
}

Application::Application(const Application &other)
{
    for (std::map<std::string, std::shared_ptr<Task>>::const_iterator it = other._tasks.begin();
         it != other._tasks.end(); ++it) {
        _tasks[it->first] = std::make_shared<Task>(*(it->second));
    }

    for (std::map<std::string, std::shared_ptr<Flow>>::const_iterator it = other._flows.begin();
         it != other._flows.end(); ++it) {
        _flows[it->first] = std::make_shared<Flow>(*(it->second));
    }
}

unsigned Application::getNumberOfTasks()
{
    return _tasks.size();
}

std::shared_ptr<Task> Application::getTask(const std::string &taskName)
{
    if (_tasks.find(taskName) != _tasks.end()) {
        return _tasks[taskName];
    } else {
        return nullptr;
    }
}

const std::map<std::string, std::shared_ptr<Task>> &Application::getTasks() const
{
    return _tasks;
}

std::vector<std::shared_ptr<Task>> Application::getTasksAsVector()
{
    std::vector<std::shared_ptr<Task>> vec;
    for (std::map<std::string, std::shared_ptr<Task>>::iterator it = _tasks.begin();
         it != _tasks.end(); ++it) {
        vec.push_back(it->second);
    }
    return vec;
}

std::vector<std::shared_ptr<Flow> > Application::getFlowsAsVector()
{
    std::vector<std::shared_ptr<Flow>> vec;
    for (std::map<std::string, std::shared_ptr<Flow>>::iterator it = _flows.begin();
         it != _flows.end(); ++it) {
        vec.push_back(it->second);
    }
    return vec;
}

unsigned Application::getNumberOfFlows()
{
    return _flows.size();
}

const std::shared_ptr<Flow> Application::getFlow(const std::string &taskName)
{
    if (_flows.find(taskName) != _flows.end()) {
        return _flows[taskName];
    } else {
        return nullptr;
    }
}

const std::map<std::string, std::shared_ptr<Flow>> &Application::getFlows() const
{
    return _flows;
}

std::string Application::getName() const
{
    std::stringstream ss(_fileName);
    std::string name;
    std::getline(ss, name, '.');
    return name;
}
