#include "processor.h"

#include <iostream>

unsigned Processor::_processorCounter = 0;

std::shared_ptr<Switcher> Processor::getSwitcher() const
{
    return _switcher;
}

const std::vector<std::shared_ptr<Task> > &Processor::getAllMappedTasks() const
{
    return _tasks;
}

Processor::Processor() :
    _id(++_processorCounter),
    _switcher(nullptr)
{
}

Processor::Processor(const Processor &other) :
    _id(++_processorCounter),
    _switcher(std::make_shared<Switcher>(*other._switcher))
{
    for (const auto &task : other._tasks) {
        _tasks.push_back(std::make_shared<Task>(*task));
    }
}

Processor::~Processor()
{
    --_processorCounter;

    for (auto &task : _tasks) {
        task = nullptr;
    }

    _switcher = nullptr;
}

bool Processor::contains(const std::shared_ptr<Task> task)
{
    bool ret = false;
    for (const std::shared_ptr<Task> mappedTask : _tasks) {
        if (mappedTask == task) {
            ret = true;
            break;
        }
    }
    return ret;
}

_float Processor::calculateUtilization()
{
    _float utilization = 0.0;
    for (std::shared_ptr<Task> task : _tasks) {
        if (task) {
            utilization += task->getCost() / task->getPeriod();
        }
    }
    return utilization;
}

void Processor::mapTask(std::shared_ptr<Task> task)
{
    _tasks.push_back(task);
}

void Processor::unmapTask(std::shared_ptr<Task> task)
{
    if (!task) {
        return;
    }

    for (std::vector<std::shared_ptr<Task>>::iterator it = _tasks.begin(); it != _tasks.end(); ++it) {
        if (*it) {
            if (task.get() == it->get()) {
                _tasks.erase(it);
                break;
            }
        }
    }
}

void Processor::mapTasks(std::vector<std::shared_ptr<Task> > tasks)
{
    _tasks.insert(_tasks.end(), tasks.begin(), tasks.end());
}

void Processor::unmapAllTasks()
{
    for (std::shared_ptr<Task> &task : _tasks) {
        task = nullptr;
    }
    _tasks.clear();
}

void Processor::connectToSwitcher(std::shared_ptr<Switcher> switcher)
{
    _switcher = switcher;
}
