#ifndef ROUTINGELEMENT_H
#define ROUTINGELEMENT_H

#include <memory>
#include <vector>

#include "../commonconfig.h"
#include "processor.h"
#include "link.h"

/*!
 * \brief The RoutingElement class is responsible for routing an appication's communication mapped
 * to this platform.
 * Performs as default the routing algorithm-XY as described in:
 * Chawade, Shubhangi D., Mahendra A. Gaikwad, and Rajendra M. Patrikar. "Review of XY routing algorithm for network-on-chip architecture." International Journal of Computer Applications 43.21 (2012): 975-8887.
 * and as used in:
 * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
 */
class RoutingElement
{
public:
    /*!
     * \brief Default Constructor.
     */
    RoutingElement();

    /*!
     * Default Destructor.
     */
    ~RoutingElement();

    /*!
     * \brief Performs routing between two processors in the platform.
     * \param links The path crossed during this routing.
     * \param processorSource The source processor.
     * \param processorDestination The destination processor.
     * \return true if the routing was successful, false otherwise.
     * \note When routing between between the same source the path crossed during the routing is empty.
     */
    bool doRouting(std::vector<std::shared_ptr<Link>> &links,
                   const std::shared_ptr<Processor> &processorSource,
                   const std::shared_ptr<Processor> &processorDestination);
};

#endif // ROUTINGELEMENT_H
