﻿/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Lucas Silveira
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <math.h>
#include <mpi.h>
#include <mutex>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <thread>

#include "costcalculator/utilizationschedulabilitycalculator.h"
#include "topology/graph.h"
#include "utilities/gerarpopuInicial.h"
#include "utilities/structure.h"

float prob_crossover;     //Crossover Probability
float porc_selection;     //% of the Population that will be selected for Crossover
float porc_reemp;           //Pct. of the Population from where it will be performed
float prob_mutation;     //Mutation Probability
int * identity;
int size;
int sentinela;



#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000




void substitution(population *pob){
    int i,j,pos_reemp;
    int conta=0;
    /*descendants replace the current Population*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        //seek position of substitution
        int tamanhoReemp = pob->topPopulation - pob->baseReplace;
        pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(values between 0 and tamanhoReemp-1)+baseReplace
        
        //apply substitution
        if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
            //copy fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copy genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
    }
    
}


void countingSort(population *pob, int k){
    int i,j;
    int *c;//c[k];
    Genome *a;
    
    c = (int*)malloc(k*sizeof(int));
    
    a = (Genome*)malloc(pob->sizePopulation*sizeof(Genome));
    if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = (int*)malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
    }
    
    /*counting sort*/
    for(i=0; i<k; i++){
        c[i] = 0;
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        c[a[i].fitness] = c[a[i].fitness] + 1;
    }
    
    for(i=1; i<k; i++){
        c[i] = c[i] + c[i-1];
    }
    
    for(i=pob->sizePopulation-1; i>=0; i--){
        
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
        c[a[i].fitness] = c[a[i].fitness] - 1;
    }
    
    
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
    free(a);
    free(c);
}


//fiz ate aqui
void mutation(population *pob,int numberprocessadores ){
    int i,j;
    float prob;
    /*mutation about descendents, baseDescendence=0*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        for(j=0; j< size; j++)
        {
            prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
            
            if(prob <= prob_mutation)
            {
                pob->genome[i].genes[j] = (rand() % numberprocessadores);
                //applied a gene mutation (flip sign)
//                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
//                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
            }
            
        }
        
    }//end for
}



//takes best solution
void selection(population *pob,  int n_generetion, int * better_soluction){
    
    countingSort(pob,pob->genome[0].size+2);
    
    if (pob->genome[0].fitness < *better_soluction)
        *better_soluction = pob->genome[0].fitness;
    
}

void crossover(population *pob) {
    int i,j,k,pos_pai1, pos_pai2, point_crossover;
    float prob;
    
    k = pob->baseDescendencia;
    
    for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
        if (k+1 == pob->sizeTotal)
            break;
        
        
        pos_pai1 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        
        for(j=0; j<10; j++){
            if (pos_pai1 != pos_pai2)
                break;
            pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        }
        
        prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
        if (prob <= prob_crossover)
        {//crossover
            point_crossover = rand() % size;//valores entre 0 e tamanho de permutation - 1
            //aplicar single_point_crossover na primera parte antes do ponto de crossover
            for(j=0; j<point_crossover; j++){
                pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            //aplicar single_point_crossover na segunda parte depois do ponto de crossover
            for(j=point_crossover; j<size; j++){
                pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            k = k + 2;
        }//end if crossover
    }//end for iniit
    pob->topeDescendencia = k;//update tope descendence
}


// Procedure that generates the initial Population







//populacao eh maior por usar apenas 24 processadores
void Population_init(population *pob, int numberprocessadores){
    int i,j,r,n,prob;
    
    //Calc size of Population
    n = size; //numero de rainhas
    pob->sizePopulation =  (int) (n * (log(n) / log(2))); //size of Population = (nlogn); log in base 2
    pob->sizePopulation =pob->sizePopulation * 24;
    pob->basePopulation = 0;
    pob->topPopulation = pob->sizePopulation;
    
    pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de Population what go to crossover
    pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a the from of where will replace
    
    pob->baseDescendencia = pob->sizePopulation;
    pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
    pob->sizeTotal = pob->topeDescendencia;
    
    pob->genome = (Genome*)malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = (int*)malloc(n * sizeof(int));//alocar memoria para cada cromosoma
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in Population_init(x,x) in Alg_Geneti_simple\n");
        pob->genome[i].size = size;
    }
    
    /*create Population a partir of  permutacion*/
    // A genom is see with a permutation of concatenation chroms
    
    for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        for (j = 0; j < pob->genome[i].size; j++) {
            int r = rand() % numberprocessadores;
            pob->genome[i].genes[j] = r;
        }
        pob->genome[i].size = size;
    }//end for extern
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
        pob->genome[i].size = size;
    }
    
    
}//end create init Population

void fitness_Population(population *pob, int flag, int numero)
{


    auto nome = std::string("App") + std::to_string(numero) + std::string(".csv");
   // printf("-------------------------------- : %d \n", size);
    std::shared_ptr<System> soc = std::make_shared<System>(std::make_shared<Application>(nome));
    soc->reconfigurePlatform(4, 6);
    static CostCalculator<int, int, _float>* calculator = new UtilizationSchedulabilityCalculator<int>(soc);
    //std::cout << "Number of Tasks: " << soc->getNumberOfApplicationTasks() << std::endl
    //          << "Number of Processors: " << soc->getNumberOfPlatformProcessors() << std::endl;


    static std::once_flag flagO;
    std::call_once(flagO, []()
    {
        calculator->setConversor([](std::vector<int> genes) -> std::vector<int>
        {
            return genes;
        });
    });

    int i;
    if (flag > 1){ //fitness over descendants
//        #pragma omp parallel for num_threads(2) private (i)
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
        {
            std::vector<int> genes(pob->genome[i].genes, pob->genome[i].genes +
                                   pob->genome[i].size);
            int eval = static_cast<int>(calculator->calculateCostFunction(genes).at(0));
            pob->genome[i].fitness = eval;
        }
    }
    else
    { // fitness over population
//        #pragma omp parallel for num_threads(2) private (i)
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
        {
            std::vector<int> genes(pob->genome[i].genes, pob->genome[i].genes +
                                   pob->genome[i].size);
            int eval = static_cast<int>(calculator->calculateCostFunction(genes).at(0));
            pob->genome[i].fitness = eval;
        }
    }
}






void Genetic(MPI_Comm communicator,int world_rank, float prob_crossover, float porcentaje_selec,float porcentaje_reemp, float prob_mutation, int appIndex, int appSize ) 
{
    permutation perm;
    sentinela=0;
    population pob;
    int better_soluction,nro_generacion,i,num_generation;
    size = appSize; //quantidade de  tarefas passadas como argumento
    perm.size=size;
    Population_init(&pob, 24);//crear poblacion inicial
    fitness_Population(&pob, 1, appIndex);
    better_soluction = pob.genome[0].fitness;
    num_generation = size;
  
    for(nro_generacion=2; nro_generacion<= size; nro_generacion++)
    { 
	   selection(&pob,nro_generacion,&better_soluction);//seleccion:sorting population
	   crossover(&pob);
	   mutation(&pob, 24);
	   fitness_Population(&pob,nro_generacion, appIndex);
	   substitution(&pob);
	
	}

	selection(&pob,nro_generacion-1,&better_soluction);
	printf("%d",better_soluction);

        for(i=0;i<pob.sizeTotal;i++){
            free(pob.genome[i].genes);
        }
        free(pob.genome);
        free(identity);
}

int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    int seed2=atoi(argv[1]);
    prob_crossover=atof(argv[2]);  //Crossover Probability
    porc_selection=atof(argv[3]);  //% of the population that will be selected for Crossover
    porc_reemp=atof(argv[4]); //Pct. of the population from where it will be performed
    prob_mutation=atof(argv[5]);      //Mutation Probability
    int appIndex=atoi(argv[6]); //nome da entrada
    int appSize=atoi(argv[7]); //tamanho da entrada
    
    int world_rank;
    MPI_Init(NULL, NULL);
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed[world_rank]+seed2);//initialize random generator
    Genetic(MPI_COMM_WORLD, world_rank, prob_crossover,porc_selection,porc_reemp, prob_mutation,appIndex, appSize);
    MPI_Finalize();
    return 0;
}


