﻿/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : GerarPopuInicial.c
 Author     : Lucas Silveira
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string>
#include "structure.h"

#include "gerarpopuInicial.h"

using namespace std;

void mostrar_poblacion(population *pob, int tamanho){
    
    int i,j;
    //imprimir poblacion actual
    for(i=pob->basePopulation; i<pob->topPopulation; i++){
        printf("(%d): ",i);
        for(j=0; j<tamanho; j++){
            printf("%d, ", pob->genome[i].genes[j]);
        }
        printf("\n");
    }
    
}

//o rank representa o processo
void CarregarPopInitial(int world_rank, float porcentaje_reemp,  float porcentaje_selec, population *pob, permutation *perm){
	int i, j=0, valor;
    FILE *arq;

    std::string fileName = "P" + std::to_string(world_rank);
    arq = fopen(fileName.c_str(), "r");
    printf("%s\n",fileName.c_str());

    if(arq == NULL)
    {
        printf("Erro, nao foi possivel abrir o arquivo\n");
        exit(0);
    }
    else
    {
        fscanf(arq,"%d", &perm->size);
        fscanf(arq,"%d", &pob->sizePopulation);//for each individual
        pob->basePopulation = 0;
        pob->topPopulation = pob->sizePopulation;
        pob->limiteSelection = pob->sizePopulation * porcentaje_selec;//limite de pob q seran "seleccionados" para crossover
        pob->baseReplace = pob->sizePopulation * porcentaje_reemp;//base de pob a partir de donde se hara el reemplazo
        
        pob->baseDescendencia = pob->sizePopulation;
        pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
        pob->sizeTotal = pob->topeDescendencia;// tamanho total = poblacion + descendencia
        /*alocar memoria*/
        pob->genome = (Genome*)malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
        if (pob->genome == NULL)
            printf("pob->genome NULL in poblacion_inicial() \n");
        for(i=0;i<pob->sizeTotal;i++){
            pob->genome[i].genes = (int*)malloc(perm->size * sizeof(int));//alocar memoria para cada cromosoma
            if (pob->genome[i].genes == NULL)
                printf("pob->genome[i].genes NULL in poblacion_inicial() \n");
        }
        
         while (j<pob->sizePopulation)
         {
             
             
             for (i=0; i<perm->size; i++)
            {
                
                fscanf(arq,"%d", &valor);
                pob->genome[j].genes[i] = valor;
            }//end for
             
            pob->genome[j].size = perm->size;
            j++;
             
             
         }//end while
        
        for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
            pob->genome[i].size = perm->size;
        }
         
         
       //  mostrar_poblacion(pob, perm->tamanho);
       
        
    }//else
    fclose(arq);
}




int * read_permutation_of_file(const char* name, permutation *perm)
{
    int i,n;
    int * identity=NULL;
    FILE *filePointer;
    if ((filePointer = fopen(name,"r")) == NULL){
        printf("qual no nome do arquivo %s\n", name);
        printf("Warning File not found\n");
    }
    
    else
    {
        //read size of permutation
        fscanf(filePointer,"%d",&n);
        perm->size = n;
        //alocate memory for perm->pi
        perm->pi = (int*)malloc(perm->size*sizeof(int));
        //leer permutacion
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
        fscanf(filePointer,"%d",&n);
        identity= (int *) calloc (n, sizeof (int));
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&identity[i]);
        }
        
        
    }//end  else
    fclose(filePointer);
    
    return identity;
}//end read permutation of archivo




//processo de debug
//Passa a quantidade de processadores a ser utilizado
void construirPopInit(int P, population *pob, int n, permutation *perm){
    int i, j, k, contador;
    i=0;
    FILE *fp;
    int conta_ind=0;
    int tamanhoPop=(int) (n * (log(n) / log(2))); //tamanho de la poblacion = (nlogn), log base 2
   
    for (contador=0; contador<P; contador++)
    {
        std::string fileName = "P" + std::to_string(contador);
        fp = fopen(fileName.c_str(), "w");
        printf("%s\n",fileName.c_str());
        
        if (fp != NULL)
        {
            fprintf(fp,"%d\n",perm->size);// tamanho da permutacao
            fprintf(fp,"%d\n",((int)(n * (log(n) / log(2)))));// tamanho da populacao
            for(k=0; k < tamanhoPop; k++)
            {
                for(j = 0; j < perm->size; j++)
                {
                    fprintf(fp,"%d\n", pob->genome[i].genes[conta_ind]);
                    conta_ind++;
                }//end for j
                i++;
                conta_ind=0;
            }//end for i1
            fclose(fp);
        }//end if
        else
        {
            printf("Falha ao criar o Arquivo");
            exit(1);
        }

    }// end for
    

        
        
    
}
