/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Lucas Angelo
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include <math.h>
#include <string.h>
#include "ordenar_pob.h"
#include "operadores.h"
#include "structs_ga.h"
#include "calc_fitness.h"
#include "GerarPopuInicial.c"

#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000

void leer_permutacion(permutacion *perm);
void slave(parameters *params);

void leer_permutacion(permutacion *perm){
	int i,n;
	scanf("%d",&n);//leer tamanho de permutacion
	perm->tamanho = n;
	for(i=0; i<n; i++){
		scanf("%d",&perm->pi[i]);
	}
}


int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, poblacion *pob, int flag){
    int i,indice;
    int porcento= sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: 
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: 
            indice=(rand() % sizePopulation);
            break;
        case 3:
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    


     if (indice > (sizePopulation-1) )
     {
         printf("ERRO AQUI NO IMIGRACAO\n\n");
     }


    return indice;
    
    
}// end function Imigracao

int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");

    switch (typeIndividual) {
        case 1:
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: 
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3:
            for (i=0; i<qtdIndividual; i++)
            {
                
                indices[i]=(rand() % sizePopulation);
                if (indices[i]>= sizePopulation)
                {
                    printf("Erro aqui no buscaIndex\n\n");
                }
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, poblacion *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    enviar.tamanho=pob->cromosomas[0].tamanho;
    enviar.fitness=pob->cromosomas[0].fitness;
    for (i=0; i<pob->cromosomas[0].tamanho; i++)
        enviar.pi[i]=pob->cromosomas[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(permutacion),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(permutacion),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->tamanhoPoblacion; j++)
                {
                    if (recebe.fitness < pob->cromosomas[j].fitness)
                    {
                        pob->cromosomas[j].fitness=recebe.fitness;
                        for (i=0; i<pob->cromosomas[j].tamanho; i++)
                        {
                            pob->cromosomas[j].genes[i]=recebe.pi[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function


int CompareIndividuals(poblacion *pob, permutacion genomaAlvo, float Pgenes) {
    int i,j;
    int flag=1;
    int contador=0;
    int intervalo=pob->tamanhoPoblacion * Pgenes;
    for (i=0; i<intervalo; i++) {
        
        for (j=0; j<pob->cromosomas[i].tamanho; j++)
        {
            if(pob->cromosomas[i].genes[j]==genomaAlvo.pi[j])
            {    contador++;
            }
        }//end for
        
        if (contador==pob->cromosomas[i].tamanho) {
            flag=0;
            break;
        }
        contador=0;
    }//end for
    
    return flag;

}


void ShareBBMMRR(MPI_Comm communicator, int world_rank, int id, poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes)
{
    int world_size,i, iii, index,j,prob;
    int flag=0;
    int avaliador=1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->tamanhoPoblacion);// funcao deve ser rescrita pra buscar uma percentual x da populacao
    for (iii=0; iii< qtdIndividual; iii++)
    {
        enviar.tamanho=pob->cromosomas[IndicesEnvia[iii]].tamanho;
        enviar.fitness=pob->cromosomas[IndicesEnvia[iii]].fitness;
        
        for (i=0; i<pob->cromosomas[0].tamanho; i++)
            enviar.pi[i]=pob->cromosomas[IndicesEnvia[iii]].genes[i];
        
        if (emigracao==2) {
            for(j = 0; j < perm->tamanho; j++){
                prob = rand() % 2;//valores entre 0 y 1
                if (prob == 1)//si prob 1, entonces gen positivo
                    pob->cromosomas[IndicesEnvia[iii]].genes[j] = perm->pi[j];
                else//si prob 0, entonces gen negativo
                    pob->cromosomas[IndicesEnvia[iii]].genes[j] = perm->pi[j] * -1;
            }//end for
            pob->cromosomas[IndicesEnvia[iii]].fitness= calcular_fitness(pob->cromosomas[IndicesEnvia[iii]], &j);// provavel dar erro
            
        }//end if emigracao ==2
        
        
        int error_code= MPI_Send(&enviar,//the data
                                     
                                     sizeof(permutacion),
                                     
                                     MPI_BYTE,
                                     
                                     id,//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            
            
        if (error_code != MPI_SUCCESS)
                
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
      
        error_code=MPI_Recv(&recebe,
                                
                                sizeof(permutacion),
                                
                                MPI_BYTE,
                                
                                 id, // processo que enviou
                                
                                0,//tag
                                
                                communicator,
                                
                                MPI_STATUS_IGNORE);
            
            
            
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
            
        if (flag)
            flag=0;
                
        else
            flag=1;
        index=Imigracao(imigracao, pob->tamanhoPoblacion, recebe.fitness,  pob,  flag);
        
        if (index>=pob->tamanhoPoblacion)
        {
            index=pob->tamanhoPoblacion-1;
            printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
        }
        avaliador=CompareIndividuals(pob, recebe, Pgenes);
         if (avaliador)
         {
             pob->cromosomas[index].fitness=recebe.fitness;
            for (i=0; i<pob->cromosomas[0].tamanho; i++)
                pob->cromosomas[index].genes[i]=recebe.pi[i];
         }
        
    }//end for

    free(IndicesEnvia);

}// end function ShareBBMMRR

void BBMMRR(MPI_Comm communicator, float *a, int *b, int tamanho,poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes){
     int i=0, error_code,world_rank,alvo=-1;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    for (; i<tamanho; i++)
    {
        if ((i%2==0 && b[i+1]==0) || (i%2==1 && b[i-1]==0) ) {
            i++;
            if (i==tamanho) {
                break;
            }
        }
        
        if (b[i]==0)
        {//recolher dados para enviar compartilhar individuos
            if (i%2==1)
                alvo=b[i-1];

            else
                alvo=b[i+1];
           
            
        }
        if (i%2==0) {
            int error_code= MPI_Send(&b[i],//the data
                                                   
                                                   sizeof(int),
                                                   
                                                   MPI_INT,
                                                   
                                                   b[i+1],//destination process
                                                   
                                                   0,//tag
                                                   
                                                   communicator);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
            }//end if
        }//end if
        else{
          //  printf("enviei o id %d para o processo %d\n", b[i], b[i-1]);
             error_code= MPI_Send(&b[i],//the data
                                     
                                     sizeof(int),
                                     
                                     MPI_INT,
                                     
                                     b[i-1],//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
            }//end if
            
        }//end else
        
    }//end for
    if (alvo==-1) {
        printf("*****************Erro o alvo nao foi setado***************\n\n");
    }
    ShareBBMMRR(communicator, world_rank, alvo, pob, perm, emigracao, qtdIndividual, typeIndividual, imigracao, Pgenes);
}//fim BBMMRR

void ParaCoordenador(MPI_Comm communicator, int world_rank, float mediaVariante, poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes){
    int world_size,iii,error_code,comp;
    Coord enviar, recebe;
    MPI_Comm_size(communicator, &world_size);
    enviar.idP=world_rank;
    enviar.MediaVariancia=mediaVariante;
    
    
    if (world_rank!=0)
        
    {
        error_code= MPI_Send(&enviar,//the data
                             sizeof(Coord),
                             MPI_BYTE,
                             0,//destination process
                             0,//tag
                             communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        error_code=MPI_Recv(&comp,
                            sizeof(int),
                            MPI_INT,
                            0, // processo que enviou
                            0,//tag
                            communicator,
                            MPI_STATUS_IGNORE);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
        ShareBBMMRR( communicator, world_rank, comp, pob, perm, emigracao, qtdIndividual, typeIndividual, imigracao, Pgenes);
    }//end if
    
    else
    {
        float retornoValue[12];
        int retornoP[12];
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(Coord),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                retornoP[iii]=recebe.idP;
                retornoValue[iii]=recebe.MediaVariancia;
            }//end if
        }//end for
        retornoP[0]=0;
        retornoValue[0]=avaliaPopulation(pob);
        MergeSort(retornoValue,retornoP, 12);
        BBMMRR(communicator,retornoValue, retornoP, world_size,pob, perm, emigracao, qtdIndividual, typeIndividual,  imigracao,  Pgenes);

    }//else
}




void mostrar_fitness(poblacion *pob, int tamanho){
    
    int i;
    for(i=pob->basePoblacion; i<pob->topePoblacion; i++){
        printf("(%d): ",i);
        
            printf("%d) %d ", i, pob->cromosomas[i].fitness);
        printf("\n");
    }
    
}




void GA(MPI_Comm communicator, int world_rank,parameters *params, float prob_crossover,float porcentaje_selec,float porcentaje_reemp,float prob_mutacion, int emigracao, int typeIndividual, int qtdIndividual,int imigracao, float intervaloMigracao)
{
    permutacion perm;
    poblacion pob, pob2;
    int mejor_solucion=-1;//mejor solucion encontrada
    int i, num_generaciones,nro_generacion, world_size;
    MPI_Comm_size(communicator, &world_size);
    leer_permutacion_de_archivo("GENOME", &perm);
    params->size_population = -1;
    params->number_generations = -1;
    params->num_eval_fit_function = -1;
    params->number_reversals = -1;
    if (world_rank==0) {
        poblacion_inicial2(&pob2,&perm,porcentaje_reemp, porcentaje_selec, world_size, params);//crear poblacion inicial
        construirPopInit(world_size,&pob2, perm.tamanho, &perm);
        for(i=0;i<pob2.tamanhoTotal;i++){
            free(pob2.cromosomas[i].genes);
        }
        free(pob2.cromosomas);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    CarregarPopInitial(world_rank, porcentaje_reemp, porcentaje_selec, &pob,&perm);
    int num_eval_fit_function = 0;
    fitness(&pob,1,&num_eval_fit_function);
    mejor_solucion = pob.cromosomas[0].fitness;
    num_generaciones = perm.tamanho;
    int contador = num_generaciones * intervaloMigracao;
    for(nro_generacion=2; nro_generacion<=num_generaciones; nro_generacion++)
    {
        
        seleccion(&pob,nro_generacion,&mejor_solucion);
        crossover3(&pob, &perm, prob_crossover);
        mutacion(&pob, &perm, prob_mutacion);
        fitness(&pob,nro_generacion,&num_eval_fit_function);
        reemplazo(&pob);
        
        contador--;
        if (contador==0)
        {
           ParaCoordenador(communicator,world_rank, avaliaPopulation(&pob),&pob, &perm,  emigracao,  qtdIndividual,  typeIndividual,  imigracao,  0.1);  
            contador = num_generaciones * intervaloMigracao;
        }
        
        if (nro_generacion==num_generaciones)
        {
            sendALL(communicator, world_rank, &pob);
            
        }//end if
        
        MPI_Barrier(MPI_COMM_WORLD);
        num_eval_fit_function=0;
    }
    
    
    if (world_rank==0)
    {    seleccion(&pob,nro_generacion,&mejor_solucion);
         printf("%d",mejor_solucion);
    }
    
    for(i=0;i<pob.tamanhoTotal;i++){
        free(pob.cromosomas[i].genes);
    }
    free(pob.cromosomas);
    
    
}

int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    unsigned int seed2 = time(NULL);
    float prob_crossover=		0.96;
    float porcentaje_selec=		0.84;
    float porcentaje_reemp=		0.60;
    float prob_mutation=		0.01;
    int emigracao=			1;
    int typeIndividual=			2;
    int qtdIndividual=			7;
    int imigracao=			2;
    float intervaloMigracao= 		0.2;
    int world_rank;
    MPI_Init(NULL, NULL);
    parameters params;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed2);
    params.seed =seed[world_rank]+seed2;
    GA(MPI_COMM_WORLD, world_rank, &params, prob_crossover,porcentaje_selec,porcentaje_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao);
    MPI_Finalize();
    return 0;
}







