/*
 ============================================================================
 Author :
 Jose Luis Soncco-Alvarez and Mauricio Ayala-Rincon
 Group of Theory of Computation
 Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include "ordenar_pob.h"
#include "structs_ga.h"

# include "mpi.h" //processamento paralelo



/*
 Dado um vetor de inteiros v e um inteiro n >= 0, ordena o vetor v[0..n-1] em ordem crescente.
 */
void MergeSort(float *v,int *v1, int n) {
    float *c = malloc(sizeof(float) * n);
    int *c1 = malloc(sizeof(int) * n);
    Sort(v,v1, c, c1, 0, n - 1);
    free(c);
}

/*
 Dado um vetor de inteiros v e dois inteiros i e f, ordena o vetor v[i..f] em ordem crescente.
 O vetor c é utilizado internamente durante a ordenação.
 */
void Sort(float *v,int *v1, float *c, int *c1, int i, int f) {
    if (i >= f) return;
    
    int m = (i + f) / 2;
    
    Sort(v, v1, c,c1, i, m);
    Sort(v, v1, c, c1, m + 1, f);
    
    /* Se v[m] <= v[m + 1], então v[i..f] já está ordenado. */
    if (v[m] <= v[m + 1]) return;
    
    Merge(v, v1, c,c1, i, m, f);
}



/*
 Dado um vetor v e três inteiros i, m e f, sendo v[i..m] e v[m+1..f] vetores ordenados,
 coloca os elementos destes vetores, em ordem crescente, no vetor em v[i..f].
 */
void Merge(float *v, int *v1, float *c, int *c1, int i, int m, int f) {
    int z,t,w,
    iv = i, ic = m + 1;
    
    for (z = i; z <= f; z++)
    {
        c[z] = v[z];
        c1[z]=v1[z];
    }
    
    z = i;
    
    while (iv <= m && ic <= f) {
        /* Invariante: v[i..z] possui os valores de v[iv..m] e v[ic..f] em ordem crescente. */
        
        if (c[iv] < c[ic])
        {
            t=z++;
            w=iv++;
            v[t] = c[w];
            v1[t]=c1[w];
        }
        else /* if (c[iv] > c[ic]) */
        {
            t=z++;
            w=ic++;
            v[t] = c[w];
            v1[t]=c1[w];
        }
    }
    
    while (iv <= m)
    {   t=z++;
        w=iv++;
        v[t] = c[w];
         v1[t] = c1[w];
    }
    
    while (ic <= f)
    {   t=z++;
        w=ic++;
        v[t] = c[w];
         v1[t] = c1[w];
    }
}





void countingSort(poblacion *pob, int k){
	int i,j;
	int c[k];
	cromosoma *a;

	/*alocar memoria para poblacion temporal "a" */
	a = malloc(pob->tamanhoPoblacion*sizeof(cromosoma));
	if (a == NULL) printf("a NULL en countingSort() de ordenar_pob.c \n");
    
    for(i=0;i<pob->tamanhoPoblacion;i++){
        a[i].genes = malloc(pob->cromosomas[0].tamanho * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL) 
            printf("a[i].genes NULL en countingSort() de ordenar_pob.c \n");
    }
    
    /*hacer una copia poblacion*/
	for(i=0; i<pob->tamanhoPoblacion; i++){
		a[i].fitness = pob->cromosomas[i].fitness;
        for(j=0; j<pob->cromosomas[0].tamanho; j++)
            a[i].genes[j] = pob->cromosomas[i].genes[j]; 
	}
    
	/*counting sort*/
	for(i=0; i<k; i++){
		c[i] = 0;
	}

	for(i=0; i<pob->tamanhoPoblacion; i++){
		c[a[i].fitness] = c[a[i].fitness] + 1;
	}

	for(i=1; i<k; i++){
		c[i] = c[i] + c[i-1];
	}

	for(i=pob->tamanhoPoblacion-1; i>=0; i--){
		//copiar en poblacion
        pob->cromosomas[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->cromosomas[0].tamanho; j++)
            pob->cromosomas[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
		c[a[i].fitness] = c[a[i].fitness] - 1;
	}
    
    
    /*Liberar memoria*/
    for(i=0;i<pob->tamanhoPoblacion;i++){
        free(a[i].genes);
    }
	free(a);
}


