/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Lucas Silveira
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include <math.h>
#include <string.h>

//#include "operadores.h"
#include "structure.h"
#include "fitness/Processing.c"
#include "GerarPopuInicial.c"
float prob_crossover;     //Crossover Probability
float porc_selection;     //% of the Population that will be selected for Crossover
float porc_reemp;           //Pct. of the Population from where it will be performed
float prob_mutation;     //Mutation Probability
int * identity;
int size;
int sentinela;



#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000




void substitution(population *pob){
    int i,j,pos_reemp;
    int conta=0;
    /*descendants replace the current Population*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        //seek position of substitution
        int tamanhoReemp = pob->topPopulation - pob->baseReplace;
        pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(values between 0 and tamanhoReemp-1)+baseReplace
        
        //apply substitution
        if (pob->genome[i].fitness < pob->genome[pos_reemp].fitness)
        {
            conta++;
            //copy fitness
            pob->genome[pos_reemp].fitness = pob->genome[i].fitness;
            //copy genes
            for(j=0; j<pob->genome[0].size; j++)
                pob->genome[pos_reemp].genes[j] = pob->genome[i].genes[j];
            
        }
    }
    
}


void countingSort(population *pob, int k){
    int i,j;
    int *c;//c[k];
    Genome *a;
    
    c = malloc(k*sizeof(int));
    
    a = malloc(pob->sizePopulation*sizeof(Genome));
    if (a == NULL) printf("a NULL en countingSort() \n");
    
    for(i=0;i<pob->sizePopulation;i++){
        a[i].genes = malloc(pob->genome[0].size * sizeof(int));//alocar memoria para cada cromosoma
        if (a[i].genes == NULL)
            printf("a[i].genes NULL en countingSort()\n");
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        a[i].fitness = pob->genome[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            a[i].genes[j] = pob->genome[i].genes[j];
    }
    
    /*counting sort*/
    for(i=0; i<k; i++){
        c[i] = 0;
    }
    
    for(i=0; i<pob->sizePopulation; i++){
        c[a[i].fitness] = c[a[i].fitness] + 1;
    }
    
    for(i=1; i<k; i++){
        c[i] = c[i] + c[i-1];
    }
    
    for(i=pob->sizePopulation-1; i>=0; i--){
        
        pob->genome[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->genome[0].size; j++)
            pob->genome[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
        c[a[i].fitness] = c[a[i].fitness] - 1;
    }
    
    
    for(i=0;i<pob->sizePopulation;i++){
        free(a[i].genes);
    }
    free(a);
    free(c);
}


//fiz ate aqui
void mutation(population *pob, permutation *perm){
    int i,j;
    float prob;
    /*mutation about descendents, baseDescendence=0*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
        for(j=0; j<perm->size; j++)
        {
            prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
            
            if(prob <= prob_mutation)
            {//applied a gene mutation (flip sign)
                if(pob->genome[i].genes[j]!=0 && j!=0 && (j+1 < size) && pob->genome[i].genes[j+1]!=0 && pob->genome[i].genes[j-1]!=0 )
                    pob->genome[i].genes[j] = pob->genome[i].genes[j] * -1;
            }
            
        }
        
    }//end for
}



//takes best solution
void selection(population *pob,  int n_generetion, int * better_soluction){
    
    countingSort(pob,pob->genome[0].size+2);
    
    if (pob->genome[0].fitness < *better_soluction)
        *better_soluction = pob->genome[0].fitness;
    
}

void crossover(population *pob, permutation *perm){
    int i,j,k,pos_pai1, pos_pai2, point_crossover;
    float prob;
    
    k = pob->baseDescendencia;
    
    for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
        if (k+1 == pob->sizeTotal)
            break;
        
        
        pos_pai1 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        
        for(j=0; j<10; j++){
            if (pos_pai1 != pos_pai2)
                break;
            pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
        }
        
        prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
        if (prob <= prob_crossover)
        {//crossover
            point_crossover = rand() % perm->size;//valores entre 0 e tamanho de permutation - 1
            //aplicar single_point_crossover na primera parte antes do ponto de crossover
            for(j=0; j<point_crossover; j++){
                pob->genome[k].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k+1].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            //aplicar single_point_crossover na segunda parte depois do ponto de crossover
            for(j=point_crossover; j<perm->size; j++){
                pob->genome[k+1].genes[j] = pob->genome[pos_pai1].genes[j];
                pob->genome[k].genes[j] = pob->genome[pos_pai2].genes[j];
            }
            k = k + 2;
        }//end if crossover
    }//end for iniit
    pob->topeDescendencia = k;//update tope descendence
}


// Procedure that generates the initial Population





void Population_init(population *pob, permutation *perm, int numberprocessadores){
    int i,j,r,n,prob;
    
    //Calc size of Population
    n = perm->size;
    pob->sizePopulation =  (int) (n * (log(n) / log(2))); //size of Population = (nlogn); log in base 2
    pob->sizePopulation=   (pob->sizePopulation*2)*numberprocessadores;
    pob->basePopulation = 0;
    pob->topPopulation = pob->sizePopulation;
    
    pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de Population what go to crossover
    pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a the from of where will replace
    
    pob->baseDescendencia = pob->sizePopulation;
    pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
    pob->sizeTotal = pob->topeDescendencia;
    
    pob->genome = malloc(pob->sizeTotal*sizeof(Genome)); //alocar memoria para poblacion total
    if (pob->genome == NULL)
        printf("pob->genome NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    
    for(i=0;i<pob->sizeTotal;i++)
    {
        pob->genome[i].genes = malloc(n * sizeof(int));//alocar memoria para cada cromosoma
        if (pob->genome[i].genes == NULL)
            printf("pob->genome[i].genes NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    }
    
    /*create Population a partir of  permutacion*/
    // A genom is see with a permutation of concatenation chroms
    
    for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
        
        for(j = 0; j < perm->size; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                r=rand();
                prob = r % 2;//values between 0 and 1
                if (prob == 1)//if prob 1, then genes is positive
                {
                    pob->genome[i].genes[j] = perm->pi[j];
                    
                }
                else//if prob 0, then genes is negative
                {
                    
                    pob->genome[i].genes[j] = perm->pi[j] * -1;
                }
            }//end if
            else
            {
                pob->genome[i].genes[j] = perm->pi[j];
            }
        }//end for internal
        
        pob->genome[i].size = perm->size;
    }//end for extern
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
        pob->genome[i].size = perm->size;
    }
    
    
}//end create init Population



//calculate the tranlocation distance to each genome of the Population
void fitness_Population(population *pob, int flag){
    int i;
    
    if (flag > 1){ //fitness over descendants
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
        {
            load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            pob->genome[i].fitness = distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            
        }
    }
    else
    { // fitness over population
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
        {
            load_genoma( pob->genome[i].genes, identity,pob->genome[i].size, size);
            pob->genome[i].fitness = distance(pob->genome[i].genes,identity, pob->genome[i].size,2);
            
        }
    }
}





//flag em uma roda vai comecar no inicio outro no final
int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, population *pob, int flag){
    int i,indice;
    int porcento=sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: // pior
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: //aleatorio
            indice=(rand() % sizePopulation);
            break;
        case 3://semelhante
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->genome[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->genome[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    
    return indice;
    
    
}// end function Imigracao


//busca os indicies a ser enviado para outra ilha
int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");
    switch (typeIndividual) {
        case 1://melhor
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: // pior
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3: //aleatorio
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=(rand() % sizePopulation);
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, population *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    msgGenome enviar,recebe;
    
    // If we are the root process, send our best data to everyone
    enviar.size=pob->genome[0].size;
    enviar.fitness=pob->genome[0].fitness;
    for (i=0; i<pob->genome[0].size; i++)
        enviar.genes[i]=pob->genome[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(msgGenome),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(msgGenome),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->sizePopulation; j++)
                {
                    if (recebe.fitness < pob->genome[j].fitness)
                    {
                        pob->genome[j].fitness=recebe.fitness;
                        for (i=0; i<pob->genome[j].size; i++)
                        {
                            pob->genome[j].genes[i]=recebe.genes[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function




//implementa topologia anel
//flag é pra marcar se for igual a 1 significa q esta terminando as geracoes e todos as ilhas necessitas saber os resultados
void my_bcastanel(MPI_Comm communicator, int world_rank, population *pob, int emigracao, int qtdIndividual, int typeIndividual, int imigracao) {
    int world_size, j,i,ii,error_code, index;
    float prob;
    int envio=-1;
    int flag=0;
    int receptor=-1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    msgGenome enviar,recebe;
    
    //buscar indices para ser enviado
    
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->sizePopulation);
    
    
    
    for (i=0; i< qtdIndividual; i++)
    {
        // If we are the root process, send our data to everyone
        enviar.size=pob->genome[IndicesEnvia[i]].size;
        enviar.fitness=pob->genome[IndicesEnvia[i]].fitness;
        for (j=0; j<pob->genome[IndicesEnvia[i]].size; j++)
            enviar.genes[j]=pob->genome[IndicesEnvia[i]].genes[j];
        
        //sent to world_rank+1
        
        if ((world_rank+1)< world_size)
        {
            envio=(world_rank+1);
            
            if (world_rank==0)
                receptor=(world_size-1);
            
            else
                receptor=(world_rank-1);
            
            
        }//end if
        
        else
        {
            envio=0;
            receptor=(world_rank-1);
            
        }
        
        
        /***************Parte onde trata emigracao segundo exclusao ************************/
        if (emigracao==2) {
            pob->genome[IndicesEnvia[i]].fitness=pob->genome[IndicesEnvia[i]].size;
        }
        
        ///sent message
        error_code= MPI_Send(&enviar,//the data
                             sizeof(msgGenome),
                             MPI_BYTE,
                             envio,//destination process
                             0,//tag
                             communicator);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
        
        
        
        error_code=MPI_Recv(&recebe,
                            sizeof(msgGenome),
                            MPI_BYTE,
                            receptor, // processo que enviou
                            0,//tag
                            communicator,
                            MPI_STATUS_IGNORE);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
        
        
        prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
        
        
        
        if (flag) {
            flag=0;
        }
        else
            flag=1;
        
        index=Imigracao( imigracao, pob->sizePopulation, recebe.fitness,  pob,  flag);
        //printf("index %d tamanho total %d\n", index,pob->sizePopulation);
        if (index>=pob->sizePopulation) {
            index=pob->sizePopulation-1;
            printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
        }
        
        pob->genome[index].fitness=recebe.fitness;
        
        for (ii=0; ii<pob->genome[index].size; ii++)
            pob->genome[index].genes[ii]=recebe.genes[ii];
        
        
        
    }//end for
    
    free(IndicesEnvia);
    
    
    
}//end function anel





void Genetic(MPI_Comm communicator,int world_rank, float prob_crossover, float porcentaje_selec,float porcentaje_reemp, float prob_mutation,int emigracao,int typeIndividual,int qtdIndividual,int imigracao,float intervaloMigracao)
{
    permutation perm;
    sentinela=0;
    population pob, pob2;
    int better_soluction,nro_generacion,i,num_generation, world_size;
    MPI_Comm_size(communicator, &world_size);
    identity=read_permutation_of_file("GENOME", &perm);
    size= perm.size;
    MPI_Barrier(MPI_COMM_WORLD);
    if (is_cotail(perm.pi,identity,perm.size, size))
    {
        
        if (world_rank==0) {
            //size=perm.size;
            Population_init(&pob2,&perm, world_size);//crear poblacion inicial
            construirPopInit(world_size,&pob2, perm.size, &perm);
            for(i=0;i<pob2.sizeTotal;i++){
               free(pob2.genome[i].genes);
            }
            free(pob2.genome);
        }
        
        MPI_Barrier(MPI_COMM_WORLD);
        
        //aqui cada processo carrega tua populacao
        CarregarPopInitial(world_rank, porcentaje_reemp, porcentaje_selec, &pob,&perm);
      //function of DadosAdicionais.c
        set_TAMANHO_CAPPEPLUS( perm.pi, perm.size);
        //function of Preprocessamento.c
        load_nodeToCompMap();
        fitness_Population(&pob, 1);
        better_soluction = pob.genome[0].fitness;
          num_generation = perm.size;
        int contador = num_generation * intervaloMigracao;
        for(nro_generacion=2; nro_generacion<=num_generation; nro_generacion++)
        {
            selection(&pob,nro_generacion,&better_soluction);//seleccion:sorting population
            crossover(&pob, &perm);
            mutation(&pob, &perm);
            fitness_Population(&pob,nro_generacion);
            substitution(&pob);
            contador--;
            if (contador==0)
            {
                //printf("entrei nesse if %d \n", world_rank);
                my_bcastanel(communicator, world_rank, &pob, emigracao, qtdIndividual,typeIndividual, imigracao);
                
                contador = num_generation * intervaloMigracao;
            }
            
            if (nro_generacion==num_generation)
            {
                sendALL(communicator, world_rank, &pob);
                
            }//end if
            
            MPI_Barrier(MPI_COMM_WORLD);
            
        }
                
        if (world_rank==0) {
            selection(&pob,nro_generacion,&better_soluction);//seleccion:ordena populacao
            printf("%d", better_soluction);
        }
        
        for(i=0;i<pob.sizeTotal;i++){
            free(pob.genome[i].genes);
        }
        free(pob.genome);
        free(identity);
        
    }
    else
        printf("*****************Input is not COTAIL************************");
   
    
    
    
}

int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    int seed2=time(NULL);
    
     prob_crossover=0.96;
     porc_selection=0.94;
     porc_reemp=0.4;
     prob_mutation=0.01;
    int emigracao=1;
    int typeIndividual=3;
    int qtdIndividual=9;
    int imigracao=1;
    float intervaloMigracao=0.1;

    
    
    int world_rank;
    MPI_Init(NULL, NULL);
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed[world_rank]+seed2);//initialize random generator
    Genetic(MPI_COMM_WORLD, world_rank, prob_crossover,porc_selection,porc_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao);
    MPI_Finalize();
    return 0;
}


